#
#  Name:    WxtImage.py
#
#  Purpose: WxPython widget to display two-dimensional 16-bit greyscale images.
#
#  Author:  William Lavender
#
#-----------------------------------------------------------------------
#
#  Copyright 2009, 2015 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

import wx
import sys

if ( wx.__version__ < "2.8" ):
	print "WxtImage requires wxPython 2.8 or above."
	sys.exit(0)

import os
import time
import math
import numpy

try:
	import Image
except:
	from PIL import Image

try:
	import ImageDraw
except:
	from PIL import ImageDraw

try:
	import ImageFont
except:
	from PIL import ImageFont

#
#=============================================================================
#
# FIXME: A better method of finding TrueType fonts must be found!
#

def Get_PIL_Font():
	fontpath = ( "/usr/share/fonts/truetype/msttcorefonts", )

	font = None

	for i in fontpath:
		path = os.path.join( i, "arial.ttf" )

		try:
			font = ImageFont.truetype( path, 14 )
		except:
			continue

		if ( font != None ):
			break

	if ( font == None ):
		font = ImageFont.load_default()

	return font

#
#=============================================================================
#

class ImagePanel(wx.Panel):

	def __init__( self, parent, size=(512,512) ):
		wx.Panel.__init__( self, parent, size=size )

		self.parent = parent

		self.rescale_factor = 1

		self.zoom_stack = []

		self.box = wx.BoxSizer( wx.VERTICAL )

		self.debug = False

		self.image_window = ImageWindow( self, -1, size=size )

		self.box.Add( self.image_window, 1, flag=wx.EXPAND )

		# Setup the border areas where the horizontal and vertical
		# scale labels will be written.

		if 0:
			self.scale_text_offset = 0
			self.scale_width = 0
		else:
			self.scale_text_offset = 5

			self.major_tick_length = 15
			self.minor_tick_length = 5

			self.font = Get_PIL_Font()

			#
			# Figure out how wide the scale labels are in units
			# of pixels.
			#

			text_size = self.font.getsize("888888")

			# print "text_size =", text_size

			self.scale_width = text_size[0] \
				+ 2 * self.scale_text_offset \
				+ self.major_tick_length
		#
		# Initialize the size of the image bitmap, so that 
		# room is left for the scales at the top and left.
		#

		window_size = self.GetSize()

		bitmap_x = window_size[0] - self.scale_width
		bitmap_y = window_size[1] - self.scale_width

		self.bitmap_size = (bitmap_x, bitmap_y)

		#----

		self.position_window = None

		self.colormap = None

		self.x_range = ( 0, size[0] - 1 )
		self.y_range = ( 0, size[1] - 1 )

		#----

		self.Bind( wx.EVT_SIZE, self.OnSize )

		if self.debug:
			print "ImagePanel: __init__ GetSize() =", self.GetSize()

	def OnSize( self, event ):
		self.Resize()

		# Update the display

		self.Display()

	def SetData( self, raw_pil_image ):

		self.raw_pil_image = raw_pil_image

		self.formatted_pil_image = None

		self.x_range = ( 0, raw_pil_image.size[0] - 1 )
		self.y_range = ( 0, raw_pil_image.size[1] - 1 )

	def GetStatistics( self ):

		start_time = time.clock()

		box = ( self.x_range[0], self.y_range[0], \
			self.x_range[1], self.y_range[1] )

		box_time = time.clock()

		cropped_pil_image = self.raw_pil_image.crop( box )

		crop_time = time.clock()

		cropped_numpy_image = numpy.asarray( cropped_pil_image )

		asarray_time = time.clock()

		mean = numpy.mean( cropped_numpy_image )

		mean_time = time.clock()

		std_dev = numpy.std( cropped_numpy_image )

		std_time = time.clock()

		min = numpy.amin( cropped_numpy_image )

		min_time = time.clock()

		max = numpy.amax( cropped_numpy_image )

		max_time = time.clock()

		if self.debug:
			print "GetStatistics():"
			print "mean =", mean, ", standard deviation =", std_dev
			print "min =", min, ", max =", max

			print "Times: box =", box_time - start_time, \
				", crop =", crop_time - box_time, \
				", asarray =", asarray_time - crop_time, \
				", mean =", mean_time - asarray_time, \
				", std =", std_time - mean_time, \
				", min =", min_time - std_time, \
				", max =", max_time - min_time, \
				", total =", max_time - start_time

		return ( mean, std_dev, min, max )

	def SetPositionWindow( self, position_window ):

		self.position_window = position_window

	def SetColormap( self, colormap ):

		self.colormap = colormap

	def SetGreyRampColormap( self, black_pixel, white_pixel ):

		if self.debug:
			print \
			"ImagePanel: SetGreyRampColormap(), black_pixel =", \
				black_pixel, ", white_pixel =", white_pixel

		if ( ( black_pixel == 0 ) and ( white_pixel == 65535 ) ):
			self.colormap = None
		else:
			x_max = 65535
			y_max = 255

			self.colormap = range( 0, x_max + 1 )

			xb = float(black_pixel)
			xw = float(white_pixel)

			# Compute the scale of the grey ramp, while
			# protecting against zero divide.

			if ( black_pixel != white_pixel ):
				scale = float(y_max) / ( xw - xb )
			else:
				scale = 0.0

			# Loop over the colormap entries.

			for x in range( 0, x_max + 1 ):

				if ( x <= black_pixel ):
					self.colormap[x] = 0
				elif ( x >= white_pixel ):
					self.colormap[x] = 255
				else:
					y = scale * ( float(x) - xb )

					# Compute the floor of y.

					y = int(y) // 1

					self.colormap[x] = y

	def apply_colormap( self ):

		if ( self.colormap == None ):
			pil_8bit_array = self.formatted_pil_image.point(
						lambda i: i * 0.00390625 )
		elif False:
			pil_8bit_array = self.apply_extremely_slow_colormap()
		else:
			pil_8bit_array = self.apply_numpy_colormap()

		return pil_8bit_array

	def apply_numpy_colormap( self ):

		start_time = time.clock()

		#
		# Create a scaled Numpy image from the scaled PIL image
		# that was created by the Resize() method.
		#

		scaled_numpy_image = \
			numpy.array( self.scaled_pil_image, dtype=numpy.uint16 )

		if self.debug:
			print "scaled_numpy_image.dtype =", \
				scaled_numpy_image.dtype
			print "scaled_numpy_image ="
			print scaled_numpy_image

		#
		# Apply the colormap to the scaled Numpy image.
		#
		# Note: This method was inspired by
		# http://mail.scipy.org/pipermail/numpy-discussion/
		#                                  2009-February/040674.html
		#

		map = self.colormap

		numpy_map = numpy.asarray( map )

		if self.debug:
			print "numpy_map ="
			print numpy_map

		recolored_32bit_numpy_image = numpy_map[ scaled_numpy_image ]

		if self.debug:
			print "recolored_32bit_numpy_image.shape =", \
				recolored_32bit_numpy_image.shape
			print "recolored_32bit_numpy_image.dtype =", \
				recolored_32bit_numpy_image.dtype
			print "recolored_32bit_numpy_image ="
			print recolored_32bit_numpy_image

		#
		# We must convert the 32bit image into an 8-bit image
		# before converting the result back to PIL format.
		#

		recolored_8bit_numpy_image = \
			recolored_32bit_numpy_image.astype( numpy.uint8 )

		if self.debug:
			print "recolored_8bit_numpy_image.shape =", \
				recolored_8bit_numpy_image.shape
			print "recolored_8bit_numpy_image.dtype =", \
				recolored_8bit_numpy_image.dtype
			print "recolored_8bit_numpy_image ="
			print recolored_8bit_numpy_image

		#
		# Convert the recolored image back to an 8-bit PIL image.
		#

		recolored_pil_image = \
			Image.fromarray( recolored_8bit_numpy_image, 'L' )

		if self.debug:
			print "recolored_pil_image.size =", \
				recolored_pil_image.size
			print "recolored_pil_image.mode =", \
				recolored_pil_image.mode
			print "recolored_pil_image ="
			print recolored_pil_image

		#
		# Apply a linear ramp to the contents of formatted_pil_image
		# so that the scale areas are correctly displayed.
		#

		new_formatted_pil_image = \
		    self.formatted_pil_image.point( lambda i: i * 0.00390625 )

		#
		# Paste the recolored image into new_formatted_pil_image.
		#

		new_formatted_pil_image.paste( recolored_pil_image,
						self.scaled_image_offset )

		end_time = time.clock()

		if self.debug:
			print "apply_numpy_colormap: total_time =", \
					end_time - start_time

		return new_formatted_pil_image

#
#  Warning: The following version of apply_colormap is TOO SLOW!
#
	def apply_extremely_slow_colormap( self ):
		# The source of image data is self.formatted_pil_image in
		# PIL 'F' format, while the result is in PIL 'L' format.

		size = self.formatted_pil_image.size

		pil_8bit_array = Image.new( 'L', size )

		src = self.formatted_pil_image.load()

		dest = pil_8bit_array.load()

		map = self.colormap

		# The parts of the image window that are used for the
		# horizontal and vertical scales do not go through the
		# colormap.  Only the image data itself, goes through
		# the colormap.

		for i in range( 0, size[0] - 1 ):
			for j in range( 0, size[1] - 1 ):
				if ( i < self.scale_width ):
					dest[i,j] = src[i,j]
				elif ( j < self.scale_width ):
					dest[i,j] = src[i,j]
				else:
					dest[i,j] = map[ int( src[i,j] ) ]

		return pil_8bit_array

	def add_pixel_text( self, pil_8bit_array ):
		if self.debug:
			print "ImagePanel: add_pixel_text(), x_range =", \
				self.x_range, ", y_range =", self.y_range

		xstart = self.x_range[0]
		xend   = self.x_range[1]
		ystart = self.y_range[0]
		yend   = self.y_range[1]

		#---

		pixel_size = self.image_to_panel_coord( \
					((xstart + 1), (ystart+1)) )

		pixel_size = ( pixel_size[0] - self.scale_width, \
				pixel_size[1] - self.scale_width )

		if self.debug:
			print "ImagePanel: add_pixel_text(), pixel_size =", \
				pixel_size

		#---

		pixel_array = self.raw_pil_image.load()

		# Copy the greyscale array to an RGB array so that we
		# can draw green text on top of it.

		pil_rgb_array = pil_8bit_array.convert("RGB")

		draw = ImageDraw.Draw( pil_rgb_array )

		# Draw the text for all the visible pixel values.

		for x in range( xstart, xend ):
			for y in range( ystart, yend ):
				image_xy = (x,y)
				panel_xy = self.image_to_panel_coord(image_xy)

				if 0:
					print "add_pixel_text(): image_xy =", \
					image_xy, ", panel_xy =", panel_xy

				pixel_value = pixel_array[x,y]

				pixel_value = int( round( pixel_value ) )

				pixel_text = str( pixel_value )

				#---

				text_size = self.font.getsize( pixel_text )

				text_offset = \
					( (pixel_size[0] - text_size[0]) // 2, \
					  (pixel_size[1] - text_size[1]) // 2 )

				panel_xy = ( panel_xy[0] + text_offset[0], \
						panel_xy[1] + text_offset[1] )

				#---

				draw.text( panel_xy, pixel_text, \
						font=self.font, fill="green" )

		return pil_rgb_array

	def Display( self ):

		if self.debug:
			print "ImagePanel: Display() invoked."

		#
		# If self.formatted_pil_image does not exist yet,
		# then just return.
		#

		try:
			test = self.formatted_pil_image
		except:
			return

		# If it is set to None, then resize the image.

		if ( self.formatted_pil_image == None ):
			self.Resize()

			self.raw_window_scale_factor = self.rescale_factor

			if self.debug:
				print \
			  "ImagePanel: Display(), raw_window_scale_factor =", \
					self.raw_window_scale_factor

		if self.debug:
			start_time = time.clock()

		# Rescale the 16-bit image to an 8-bit image (still stored
		# in 32-bit 'I' format at this point).

		pil_8bit_array = self.apply_colormap()

		if self.debug:
			convert_to_8bit_time = time.clock()

		# print "self.rescale_factor =", self.rescale_factor

		if ( self.rescale_factor >= 50.0 ):
			pil_8bit_array = self.add_pixel_text( pil_8bit_array )

		if self.debug:
			add_pixel_text_time = time.clock()

		# Create a wx.Image from the 8-bit PIL image.

		wx_image = wx.EmptyImage( pil_8bit_array.size[0], \
					pil_8bit_array.size[1] )

		wx_image.SetData( pil_8bit_array.convert("RGB").tostring() )

		if self.debug:
			convert_to_wx_time = time.clock()

		self.image_window.Display( wx_image )

		if self.debug:
			copy_time = time.clock()

			print \
"ImagePanel.Display(): convert to 8bit = %f, add pixel text = %f, convert to wx = %f, copy = %f" \
				% ( convert_to_8bit_time - start_time, \
				add_pixel_text_time - convert_to_8bit_time, \
				convert_to_wx_time - add_pixel_text_time, \
				copy_time - convert_to_wx_time )

	def Resize( self ):
		window_size = self.GetSize()

		if self.debug:
			print "ImagePanel: Resize() size =", window_size

		bitmap_x = window_size[0] - self.scale_width
		bitmap_y = window_size[1] - self.scale_width

		self.bitmap_size = (bitmap_x, bitmap_y)

		if self.debug:
			print "self.bitmap_size =", self.bitmap_size

		# Compute the width and height of the zoom region.

		try:
			int_x = self.x_range[1] - self.x_range[0]
			int_y = self.y_range[1] - self.y_range[0]
		except:
			return

		if self.debug:
			print "int_x =", int_x, ", int_y =", int_y

		# If the user just clicked on the window to raise it to
		# the top of the window stack, then probably one or both
		# of int_x and int_y will be zero.  If so, then we just
		# return without doing anything.

#		if ( ( int_x == 0 ) or ( int_y == 0 ) ):
#			return

		# Tell the windowing system to redraw the entire contents
		# of this window.

		self.Refresh()

		# Compute a rescale factor that will be used to resize the
		# raw_pil_array to match the current size of the panel.

		raw_x = float( int_x )
		raw_y = float( int_y )

		if self.debug:
			print "raw_x =", raw_x, ", raw_y =", raw_y

		x_ratio = float( bitmap_x ) / raw_x
		y_ratio = float( bitmap_y ) / raw_y

		if ( x_ratio <= y_ratio ):
			self.rescale_factor = x_ratio
		else:
			self.rescale_factor = y_ratio

		if self.debug:
			print "ImagePanel: Resize(): rescale_factor =", \
				self.rescale_factor

		if ( self.rescale_factor <= 0.0 ):
			if self.debug:
				print "rescale_factor is <= 0.  Giving up."
			return

		new_x = int( self.rescale_factor * raw_x )
		new_y = int( self.rescale_factor * raw_y )

		new_size = ( new_x, new_y )

		if self.debug:
			print "ImagePanel: Resize(): new_size =", new_size

		# Compute the position in panel units of the lower right
		# hand corner of the image.  The clipped_panel_coord()
		# method will use this to prevent clipped panel coordinates
		# from falling outside the image window.

		new_panel_limits = self.image_to_panel_coord( \
					( self.x_range[1], self.y_range[1] ) )

		if self.debug:
			print "new_panel_limits =", new_panel_limits

		self.max_panel_coord = new_panel_limits

		# Construct a cropped PIL image that contains only pixels
		# within self.x_range and self.y_range

		box = ( self.x_range[0], self.y_range[0], \
			self.x_range[1], self.y_range[1] )

		try:
			cropped_pil_image = self.raw_pil_image.crop( box )
		except:
			# self.raw_pil_image does not exist yet.

			return

		if self.debug:
			print "Created cropped PIL image"

		# Now resize the cropped PIL image to a scaled PIL image.

		self.scaled_pil_image = cropped_pil_image.resize( new_size )

		if self.debug:
			print "Created scaled PIL image"

		# If we are not displaying the horizontal and vertical scale
		# labels, then we are done.

		if ( self.scale_width == 0 ):
			self.formatted_pil_image = self.scaled_pil_image
			return

		# Create an empty PIL image that is the full size of the
		# panel.  We will build here the final image to be written
		# to the display.

		background = 60000.0

		self.formatted_pil_image = Image.new( "F", \
						self.GetSize(), background )

		# Copy in the scaled image.

		self.scaled_image_offset = \
			( self.scale_width, self.scale_width )

		self.formatted_pil_image.paste( self.scaled_pil_image, \
						self.scaled_image_offset )

		# Create an ImageDraw object to draw the tick marks on.

		draw = ImageDraw.Draw( self.formatted_pil_image )

		# Add the scale ticks.

		self.add_ticks( draw, 'X' )

		self.add_ticks( draw, 'Y' )

	def Zoom( self, x_range, y_range ):

		# Save the old zoom settings.

		self.zoom_stack.append( (self.x_range, self.y_range) )

		if self.debug:
			print "ImagePanel: Zoom() zoom_stack =", self.zoom_stack

		# Zoom the image.

		self.x_range = x_range
		self.y_range = y_range

		self.Resize()

		self.Display()

	def UnZoom( self ):

		if ( len(self.zoom_stack) == 0 ):
			if self.debug:
				print "ImagePanel: UnZoom(): nothing to unzoom."
			return

		( x_range, y_range ) = self.zoom_stack.pop()

		if self.debug:
			print "ImagePanel: UnZoom(), x_range =", x_range, \
					", y_range =", y_range

			print "ImagePanel: UnZoom() zoom_stack =", \
					self.zoom_stack

		# Unzoom the image.

		self.x_range = x_range
		self.y_range = y_range

		self.Resize()

		self.Display()

	def compute_tick_step_sizes( self, range_length ):

		# print "compute_step: range_length =", range_length

		log_range = math.log10( range_length )

		# print "compute_step: log_range =", log_range

		base = int(log_range)
		base = math.pow( 10.0, base )

		offset = log_range - int(log_range)

		# print "compute_step: base =", base, ", offset =", offset

		if ( offset < 0.25 ):
			major = 0.2 * base
			minor = 0.05 * base

		else:
			major = 0.5 * base
			minor = 0.1 * base

		# print "compute_step: major =", major, ", minor =", minor

		if ( major < 1.0 ):
			major = 1
		if ( minor < 1.0 ):
			minor = 1

		major = round( major )
		minor = round( minor )

		return ( major, minor )

	def add_ticks( self, draw, axis ):
		if ( axis == 'X' ):
			range_start = self.x_range[0]
			range_end = self.x_range[1]
		else:
			range_start = self.y_range[0]
			range_end = self.y_range[1]

		range_length = range_end - range_start + 1

		if self.debug:
			print "add_ticks: %s start = %i, end = %i" \
				%( axis, range_start, range_end )

		( tick_major_step_size, tick_minor_step_size ) \
			= self.compute_tick_step_sizes( range_length )

		if self.debug:
			print "add_ticks: %s major step = %i, minor step = %i" \
			  % ( axis, tick_major_step_size, tick_minor_step_size )

		# Here we use integer division.

		tick = \
		  ( range_start // tick_minor_step_size ) * tick_minor_step_size

		if ( tick < range_start ):
			tick = tick + tick_minor_step_size

		while ( True ):

			if ( (tick % tick_major_step_size) == 0 ):
				self.add_major_tick( draw, axis, tick )
			else:
				self.add_minor_tick( draw, axis, tick )

			tick = tick + tick_minor_step_size

			if ( tick > range_end ):
				break
			
	def add_major_tick( self, draw, axis, tick ):

		tick_centerline = self.get_tick_centerline( axis, tick )

		self.add_tick_mark( draw, axis, tick_centerline, \
					self.major_tick_length )

		tick_string = str(tick)

		string_size = self.font.getsize(tick_string)

		if ( axis == 'X' ):
			x = tick_centerline - (string_size[0] // 2)

			y = self.scale_width - self.scale_text_offset \
				- string_size[1] - self.major_tick_length
		else:
			x = self.scale_width - self.scale_text_offset \
				- string_size[0] - self.major_tick_length

			y = tick_centerline - (string_size[1] // 2)

		draw.text( (x,y), tick_string, font=self.font )

	def add_minor_tick( self, draw, axis, tick ):

		tick_centerline = self.get_tick_centerline( axis, tick )

		self.add_tick_mark( draw, axis, tick_centerline, \
					self.minor_tick_length )

	def get_tick_centerline( self, axis, tick ):
		
		if ( axis == 'X' ):
			range_start = self.x_range[0]
		else:
			range_start = self.y_range[0]

		range_offset = tick - range_start

		panel_offset = self.scale_width \
			+ int( round( range_offset * self.rescale_factor ) )

		return panel_offset

	def add_tick_mark( self, draw, axis, tick_centerline, tick_length ):

		if ( axis == 'X' ):
			tick_mark_y_end   = self.scale_width
			tick_mark_y_start = tick_mark_y_end - tick_length

			tick_mark_x = tick_centerline 

			line = ( (tick_mark_x, tick_mark_y_start), \
				(tick_mark_x, tick_mark_y_end) )
		else:
			tick_mark_x_end   = self.scale_width
			tick_mark_x_start = tick_mark_x_end - tick_length

			tick_mark_y = tick_centerline

			line = ( (tick_mark_x_start, tick_mark_y), \
				(tick_mark_x_end, tick_mark_y) )

		draw.line( line )

	def image_to_panel_coord( self, image_coord ):

		image_x = image_coord[0] - self.x_range[0]
		image_y = image_coord[1] - self.y_range[0]

		panel_x = self.scale_width \
			+ int( round( image_x * self.rescale_factor ) )

		panel_y = self.scale_width \
			+ int( round( image_y * self.rescale_factor ) )

		return [panel_x, panel_y]

	def panel_to_image_coord( self, panel_coord ):

		panel_x = panel_coord[0]
		panel_y = panel_coord[1]

		x_max = self.x_range[1]
		y_max = self.y_range[1]

		image_x = self.x_range[0] + \
	    int( round( (panel_x - self.scale_width) / self.rescale_factor ) )

		if ( image_x < 0 ):
			return None
		if ( image_x > x_max ):
			return None

		image_y = self.y_range[0] + \
	    int( round( (panel_y - self.scale_width) / self.rescale_factor ) )

		if ( image_y < 0 ):
			return None
		if ( image_y > y_max ):
			return None

		return [image_x, image_y]

	def panel_to_clipped_image_coord( self, panel_coord ):

		panel_x = panel_coord[0]
		panel_y = panel_coord[1]

		if ( self.debug ):
			print "panel_to_clipped_image_coord() invoked."
			print "panel_x =", panel_x, ", panel_y =", panel_y

		x_max = self.x_range[1]
		y_max = self.y_range[1]

		if ( self.debug ):
			print "x_max =", x_max, ", y_max =", y_max

		image_x = self.x_range[0] + \
	    int( round( (panel_x - self.scale_width) / self.rescale_factor ) )

		if ( self.debug ):
			print "original image_x =", image_x

		if ( image_x < 0 ):
			image_x = 0
		if ( image_x > x_max ):
			image_x = x_max

		image_y = self.y_range[0] + \
	    int( round( (panel_y - self.scale_width) / self.rescale_factor ) )

		if ( self.debug ):
			print "original image_y =", image_y

		if ( image_y < 0 ):
			image_y = 0
		if ( image_y > y_max ):
			image_y = y_max

		if ( self.debug ):
			print "image_x =", image_x, ", image_y =", image_y
			print "panel_to_clipped_image_coord() complete."

		return [image_x, image_y]

	def clipped_panel_coord( self, panel_coord ):

		panel_x = panel_coord[0]
		panel_y = panel_coord[1]

		panel_size = self.GetSize()

		if ( panel_x < self.scale_width ):
			panel_x = self.scale_width

		if ( panel_y < self.scale_width ):
			panel_y = self.scale_width

		if ( panel_x >= self.max_panel_coord[0] ):
			panel_x = self.max_panel_coord[0] - 1

		if ( panel_y >= self.max_panel_coord[1] ):
			panel_y = self.max_panel_coord[1] - 1

		return [panel_x, panel_y]

	def get_pixel_size( self ):

		xstart = self.x_range[0]
		ystart = self.y_range[0]

		if ( xstart == 0 ):
			xstart = self.scale_width

		if ( ystart == 0 ):
			ystart = self.scale_width

		pixel_size = self.image_to_panel_coord( \
						((xstart + 1), (ystart + 1)) )

		pixel_size = ( pixel_size[0] - self.scale_width, \
				pixel_size[1] - self.scale_width )

		return pixel_size

#
#=============================================================================
#

class ImageWindow(wx.Window):

	def __init__( self, parent, id=-1, size=(50,50) ):
		wx.Window.__init__( self, parent, id=id, size=size )

		self.debug = False

		self.parent = parent

		self.skip_event = True

		self.num_fast_onpaint_exceptions = 0

		self.last_onpaint_exception_time = 0

		self.SetBackgroundColour("skyblue")

		if self.debug:
			print "ImageWindow: __init__ GetSize() =", \
						self.GetSize()

		self.refresh_display = False

		self.Bind( wx.EVT_IDLE, self.OnIdle )
		self.Bind( wx.EVT_PAINT, self.OnPaint )

		self.Bind( wx.EVT_LEFT_DOWN, self.OnLeftDown )
		self.Bind( wx.EVT_MOTION, self.OnMotion )
		self.Bind( wx.EVT_LEFT_UP, self.OnLeftUp )

		self.Bind( wx.EVT_RIGHT_UP, self.OnRightUp )

		self.start_position = None

		self.overlay = wx.Overlay()

	def Display( self, wx_image ):
		if self.debug:
			print "ImageWindow: Display()"

		self.bitmap = wx.BitmapFromImage( wx_image )

		self.refresh_display = True

	def ResetOverlay( self ):
		# Reset the overlay, which restores the original contents
		# of the window.

		dc = wx.ClientDC( self )

		dc_overlay = wx.DCOverlay( self.overlay, dc )
		dc_overlay.Clear()
		del dc_overlay

		self.overlay.Reset()

	def OnIdle( self, event ):

		if self.refresh_display:
			if self.debug:
				print "ImageWindow: OnIdle()"

			self.refresh_display = False

			dc = wx.BufferedDC( wx.ClientDC(self), self.bitmap )

	def OnPaint( self, event ):
		if 1 and self.debug:
			print "ImageWindow: OnPaint()"

		try:
			dc = wx.BufferedPaintDC( self, self.bitmap )
		except:
			print "ImageWindow: OnPaint() exception."

			if self.skip_event:
				self.skip_event = False
				print "ImageWindow: OnPaint() event skipped."
				event.Skip()

				return

			current_time = time.time()

			diff = current_time - self.last_onpaint_exception_time

			if ( diff < 0.1 ):
				self.num_fast_onpaint_exceptions += 1
			else:
				self.num_fast_onpaint_exceptions = 0

			self.last_onpaint_exception_time = current_time

			if ( self.num_fast_onpaint_exceptions > 100 ):
				print \
				"Exiting due to too many OnPaint() exceptions."

				sys.exit(1)

	def OnLeftDown( self, event ):
		panel = self.parent

		# Capture the mouse and save the starting position for
		# the rubber band.

		self.CaptureMouse()

		panel_position = event.GetPosition()

		self.panel_start_position = \
			panel.clipped_panel_coord( panel_position )

		if self.debug:
			print \
			"ImageWindow: OnLeftDown(), panel_start_position =", \
				self.panel_start_position

	def OnMotion( self, event ):
		panel = self.parent

		panel_position = event.GetPosition()

		if False:
			image_position = \
				panel.panel_to_image_coord( panel_position )

			print "ImageWindow: OnMotion() at", panel_position, \
				", image =", image_position

		if ( panel.position_window != None ):

			# panel_to_image_coord() returns an image location
			# for the center of the pixel.  For this case, we
			# want the image location to be attached to the
			# upper left corner of the pixel.

			pixel_size = panel.get_pixel_size()

			panel_position[0] = \
		    round( panel_position[0] - 0.5 * float(pixel_size[0]) )

			panel_position[1] = \
		    round( panel_position[1] - 0.5 * float(pixel_size[1]) )

			image_position = \
				panel.panel_to_image_coord( panel_position )

			if ( image_position != None ):

				x = image_position[0]
				y = image_position[1]

				try:
					raw_pil_image = panel.raw_pil_image

					pixel_array = raw_pil_image.load()

					pixel_value = pixel_array[x,y]

					pixel_value = \
						int( round( pixel_value ) )

					text = "X = %d, Y = %d,  Value = %d" \
						% ( x, y, pixel_value )

					panel.position_window.SetLabel( text )
				except:
					pass

		if ( event.Dragging() and event.LeftIsDown() ):

			panel_position = \
				panel.clipped_panel_coord( panel_position )

			rect = wx.RectPP( self.panel_start_position, \
						panel_position )

			# Draw the rectangle on a DC overlay.  This keeps
			# the rectangle separate from the original window
			# contents.

			dc = wx.ClientDC( self )

			dc_overlay = wx.DCOverlay( self.overlay, dc )
			dc_overlay.Clear()

			dc.SetPen( wx.Pen("red", 2) )

			if "wxMac" in wx.PlatformInfo:
				dc.SetBrush( wx.Brush( \
					wx.Colour( 0xC0, 0xC0, 0xC0, 0x80) ) )
			else:
				dc.SetBrush( wx.TRANSPARENT_BRUSH )

			dc.DrawRectangleRect( rect )

			# Apparently we must make sure that the DC overlay
			# is destroyed before the DC is.

			del dc_overlay

	def OnLeftUp( self, event ):
		panel = self.parent

		if self.HasCapture():
			self.ReleaseMouse()

		panel_finish_position = event.GetPosition()

		panel_finish_position = \
			panel.clipped_panel_coord( panel_finish_position )

		if self.debug:
			print \
			"ImageWindow: OnLeftUp(), self.panel_start_position =",\
				self.panel_start_position
			print \
			"ImageWindow: OnLeftUp(), panel_finish_position =",\
				panel_finish_position

		# Reset the overlay, which restores the original contents 
		# of the window.

		self.ResetOverlay()

		# Compute the image coordinates for the new zoomed image.

		image_start = panel.panel_to_clipped_image_coord( \
					self.panel_start_position )

		image_finish = panel.panel_to_clipped_image_coord( \
					panel_finish_position )

		if self.debug:
			print \
			"ImageWindow: OnLeftUp(), image_start =", image_start,\
				", image_finish =", image_finish

		# If the user clicked on our window to raise it to the top
		# of the window stack, but not to zoom it, then the zoom
		# window will have either a zero height or a zero width
		# or both.  If this has happened, then we do not want to
		# actually zoom.

		if ( image_start[0] == image_finish[0] ):
			if self.debug:
				print \
				"ImageWindow: OnLeftUp() zoom width is zero."
			return

		if ( image_start[1] == image_finish[1] ):
			if self.debug:
				print \
				"ImageWindow: OnLeftUp() zoom height is zero."
			return

		#
		# Get the panel pixel size.
		#

		pixel_size = panel.get_pixel_size()

		raw_image_size = panel.raw_pil_image.size

		if self.debug:
			print "ImageWindow: pixel_size =", pixel_size
			print "ImageWindow: raw_image_size =", raw_image_size

		#
		# panel_to_clipped_image_coord() gives us pixel values that
		# correspond to the upper left corner of the pixel.  If the
		# values of image_finish[0] or image_finish[1] are closer
		# to the corresponding edge of the unscaled image, then we
		# _really_ want the lower side or the right side of the
		# pixel, or both.
		#
		# For example, suppose we have a 4096 by 4096 image, with an
		# X-pixel size of 14, and image_start[0] < image_finish[0],
		# Then, if image_finish[0] = 4090, then image_finish[0]
		# will be set to 4096, since the difference is less than 14.
		#
		# Otherwise, if image_finish[0] = 3789, then image_finish[0]
		# will be left alone.
		#

		# Do the X-axis first.

		if ( image_start[0] < image_finish[0] ):
			diff = raw_image_size[0] - image_finish[0]

			if ( diff < pixel_size[0] ):
				image_finish[0] = raw_image_size[0]
		else:
			diff = raw_image_size[0] - image_start[0]

			if ( diff < pixel_size[0] ):
				image_start[0] = raw_image_size[0]

		# Then do the Y-axis.

		if ( image_start[1] < image_finish[1] ):
			diff = raw_image_size[1] - image_finish[1]

			if ( diff < pixel_size[1] ):
				image_finish[1] = raw_image_size[1]
		else:
			diff = raw_image_size[1] - image_start[1]

			if ( diff < pixel_size[1] ):
				image_start[1] = raw_image_size[1]

		if self.debug:
			print \
		    "ImageWindow before(zoom) : image_start =", image_start, \
				", image_finish =", image_finish

		#
		# Compute the zoom range.
		#

		if ( image_start[0] < image_finish[0] ):
			x_range = ( image_start[0], image_finish[0] )
		else:
			x_range = ( image_finish[0], image_start[0] )
		
		if ( image_start[1] < image_finish[1] ):
			y_range = ( image_start[1], image_finish[1] )
		else:
			y_range = ( image_finish[1], image_start[1] )
		
		# Zoom the image.

		panel.Zoom( x_range, y_range )

		#

		self.panel_start_position = None

	def OnRightUp( self, event ):
		panel = self.parent

		if self.debug:
			print "ImageWindow: OnRightUp()"

		panel.UnZoom()

